@@ -1,4 +1,4 @@
-// Copyright (c) 2018-2021  Brendan Molloy <brendan@bbqsrc.net>,
+// Copyright (c) 2018-2022  Brendan Molloy <brendan@bbqsrc.net>,
 //                          Ilya Solovyiov <ilya.solovyiov@gmail.com>,
 //                          Kai Ren <tyranron@gmail.com>
 //
@@ -34,7 +34,6 @@ use futures::{
 };
 use itertools::Itertools as _;
 use regex::{CaptureLocations, Regex};
-use structopt::StructOpt;
 
 use crate::{
     event::{self, HookType, Info},
@@ -42,18 +41,12 @@ use crate::{
     parser, step, Event, Runner, Step, World,
 };
 
-// Workaround for overwritten doc-comments.
-// https://github.com/TeXitoi/structopt/issues/333#issuecomment-712265332
-#[cfg_attr(doc, doc = "CLI options of a [`Basic`] [`Runner`].")]
-#[cfg_attr(
-    not(doc),
-    allow(clippy::missing_docs_in_private_items, missing_docs)
-)]
-#[derive(Clone, Copy, Debug, StructOpt)]
+/// CLI options of a [`Basic`] [`Runner`].
+#[derive(clap::Args, Clone, Copy, Debug)]
 pub struct Cli {
     /// Number of scenarios to run concurrently. If not specified, uses the
     /// value configured in tests runner, or 64 by default.
-    #[structopt(long, short, name = "int")]
+    #[clap(long, short, name = "int")]
     pub concurrency: Option<usize>,
 }
 
@@ -502,7 +495,7 @@ async fn execute<W, Before, After>(
     // 1. We obtain the current panic hook and replace it with an empty one.
     // 2. We run tests, which can panic. In that case we pass all panic info
     //    down the line to the Writer, which will print it at a right time.
-    // 3. We return original panic hook, because suppressing all panics doesn't
+    // 3. We restore original panic hook, because suppressing all panics doesn't
     //    sound like a very good idea.
     let hook = panic::take_hook();
     panic::set_hook(Box::new(|_| {}));
@@ -708,8 +701,8 @@ where
 
             let rule_background = rule
                 .as_ref()
-                .map(|rule| {
-                    rule.background
+                .map(|r| {
+                    r.background
                         .as_ref()
                         .map(|b| b.steps.iter().map(|s| Arc::new(s.clone())))
                         .into_iter()
@@ -861,6 +854,7 @@ where
                 }
             };
 
+            #[allow(clippy::shadow_unrelated)]
             match fut.await {
                 Ok(world) => {
                     self.send(event::Cucumber::scenario(
@@ -956,6 +950,7 @@ where
             }
         };
 
+        #[allow(clippy::shadow_unrelated)]
         match run.await {
             Ok((Some(captures), Some(world))) => {
                 self.send(passed(step, captures));
@@ -1047,16 +1042,18 @@ where
         }
 
         let mut started_rules = Vec::new();
-        for (feature, rule) in runnable
+        for (feat, rule) in runnable
             .iter()
-            .filter_map(|(f, r, _)| r.clone().map(|r| (Arc::clone(f), r)))
+            .filter_map(|(feat, rule, _)| {
+                rule.clone().map(|r| (Arc::clone(feat), r))
+            })
             .dedup()
         {
             let _ = self
                 .rule_scenarios_count
-                .entry((feature.path.clone(), Arc::clone(&rule)))
+                .entry((feat.path.clone(), Arc::clone(&rule)))
                 .or_insert_with(|| {
-                    started_rules.push((feature, rule));
+                    started_rules.push((feat, rule));
                     0.into()
                 });
         }
@@ -1172,11 +1169,11 @@ impl Features {
                     .map(|s| (&feature, Some(r), s))
                     .collect::<Vec<_>>()
             }))
-            .map(|(f, r, s)| {
+            .map(|(feat, rule, scenario)| {
                 (
-                    Arc::new(f.clone()),
-                    r.map(|r| Arc::new(r.clone())),
-                    Arc::new(s.clone()),
+                    Arc::new(feat.clone()),
+                    rule.map(|r| Arc::new(r.clone())),
+                    Arc::new(scenario.clone()),
                 )
             })
             .into_group_map_by(|(f, r, s)| {
