@@ -1,4 +1,4 @@
-// Copyright (c) 2018-2021  Brendan Molloy <brendan@bbqsrc.net>,
+// Copyright (c) 2018-2022  Brendan Molloy <brendan@bbqsrc.net>,
 //                          Ilya Solovyiov <ilya.solovyiov@gmail.com>,
 //                          Kai Ren <tyranron@gmail.com>
 //
@@ -15,42 +15,47 @@ use std::{
     cmp,
     fmt::{Debug, Display},
     io,
-    ops::Deref,
     str::FromStr,
 };
 
 use async_trait::async_trait;
-use console::Term;
+use derive_more::{Deref, DerefMut};
 use itertools::Itertools as _;
 use regex::CaptureLocations;
-use structopt::StructOpt;
 
 use crate::{
+    cli::Colored,
     event::{self, Info},
     parser,
-    writer::term::Styles,
-    ArbitraryWriter, Event, World, Writer,
+    writer::{
+        self,
+        out::{Styles, WriteStrExt as _},
+        Ext as _, Verbosity,
+    },
+    Event, World, Writer,
 };
 
-// Workaround for overwritten doc-comments.
-// https://github.com/TeXitoi/structopt/issues/333#issuecomment-712265332
-#[cfg_attr(doc, doc = "CLI options of a [`Basic`] [`Writer`].")]
-#[cfg_attr(
-    not(doc),
-    allow(missing_docs, clippy::missing_docs_in_private_items)
-)]
-#[derive(Clone, Copy, Debug, StructOpt)]
+/// CLI options of a [`Basic`] [`Writer`].
+#[derive(clap::Args, Clone, Copy, Debug)]
 pub struct Cli {
-    /// Increased verbosity of an output: additionally outputs step's doc
-    /// string (if present).
-    #[structopt(long, short)]
-    pub verbose: bool,
+    /// Verbosity of an output.
+    ///
+    /// `-v` is default verbosity, `-vv` additionally outputs world on failed
+    /// steps, `-vvv` additionally outputs step's doc string (if present).
+    #[clap(short, parse(from_occurrences))]
+    pub verbose: u8,
 
     /// Coloring policy for a console output.
-    #[structopt(long, name = "auto|always|never", default_value = "auto")]
+    #[clap(long, name = "auto|always|never", default_value = "auto")]
     pub color: Coloring,
 }
 
+impl Colored for Cli {
+    fn coloring(&self) -> Coloring {
+        self.color
+    }
+}
+
 /// Possible policies of a [`console`] output coloring.
 #[derive(Clone, Copy, Debug)]
 pub enum Coloring {
@@ -78,15 +83,27 @@ impl FromStr for Coloring {
     }
 }
 
-/// Default [`Writer`] implementation outputting to [`Term`]inal (STDOUT by
-/// default).
+/// Default [`Writer`] implementation outputting to an [`io::Write`] implementor
+/// ([`io::Stdout`] by default).
 ///
 /// Pretty-prints with colors if terminal was successfully detected, otherwise
 /// has simple output. Useful for running tests with CI tools.
-#[derive(Debug)]
-pub struct Basic {
-    /// Terminal to write the output into.
-    terminal: Term,
+///
+/// # Ordering
+///
+/// This [`Writer`] isn't [`Normalized`] by itself, so should be wrapped into
+/// a [`writer::Normalize`], otherwise will produce output [`Event`]s in a
+/// broken order.
+///
+/// [`Normalized`]: writer::Normalized
+/// [`Runner`]: crate::runner::Runner
+/// [`Scenario`]: gherkin::Scenario
+#[derive(Debug, Deref, DerefMut)]
+pub struct Basic<Out: io::Write = io::Stdout> {
+    /// [`io::Write`] implementor to write the output into.
+    #[deref]
+    #[deref_mut]
+    output: Out,
 
     /// [`Styles`] for terminal output.
     styles: Styles,
@@ -96,33 +113,36 @@ pub struct Basic {
 
     /// Number of lines to clear.
     lines_to_clear: usize,
+
+    /// [`Verbosity`] of this [`Writer`].
+    verbosity: Verbosity,
 }
 
 #[async_trait(?Send)]
-impl<W: World + Debug> Writer<W> for Basic {
+impl<W, Out> Writer<W> for Basic<Out>
+where
+    W: World + Debug,
+    Out: io::Write,
+{
     type Cli = Cli;
 
     #[allow(clippy::unused_async)] // false positive: #[async_trait]
     async fn handle_event(
         &mut self,
         ev: parser::Result<Event<event::Cucumber<W>>>,
-        cli: &Self::Cli,
+        opts: &Self::Cli,
     ) {
         use event::{Cucumber, Feature};
 
-        match cli.color {
-            Coloring::Always => self.styles.is_present = true,
-            Coloring::Never => self.styles.is_present = false,
-            Coloring::Auto => {}
-        };
+        self.apply_cli(*opts);
 
         match ev.map(Event::into_inner) {
             Err(err) => self.parsing_failed(&err),
             Ok(Cucumber::Started | Cucumber::Finished) => Ok(()),
             Ok(Cucumber::Feature(f, ev)) => match ev {
                 Feature::Started => self.feature_started(&f),
-                Feature::Scenario(sc, ev) => self.scenario(&f, &sc, &ev, *cli),
-                Feature::Rule(r, ev) => self.rule(&f, &r, ev, *cli),
+                Feature::Scenario(sc, ev) => self.scenario(&f, &sc, &ev),
+                Feature::Rule(r, ev) => self.rule(&f, &r, ev),
                 Feature::Finished => Ok(()),
             },
         }
@@ -131,10 +151,11 @@ impl<W: World + Debug> Writer<W> for Basic {
 }
 
 #[async_trait(?Send)]
-impl<'val, W, Val> ArbitraryWriter<'val, W, Val> for Basic
+impl<'val, W, Val, Out> writer::Arbitrary<'val, W, Val> for Basic<Out>
 where
     W: World + Debug,
     Val: AsRef<str> + 'val,
+    Out: io::Write,
 {
     #[allow(clippy::unused_async)] // false positive: #[async_trait]
     async fn write(&mut self, val: Val)
@@ -142,40 +163,80 @@ where
         'val: 'async_trait,
     {
         self.write_line(val.as_ref())
-            .unwrap_or_else(|e| panic!("Failed to write into terminal: {}", e));
+            .unwrap_or_else(|e| panic!("Failed to write: {}", e));
     }
 }
 
-impl Default for Basic {
-    fn default() -> Self {
-        Self {
-            terminal: Term::stdout(),
+impl<O: io::Write> writer::NonTransforming for Basic<O> {}
+
+impl Basic {
+    /// Creates a new [`Normalized`] [`Basic`] [`Writer`] outputting to
+    /// [`io::Stdout`].
+    ///
+    /// [`Normalized`]: writer::Normalized
+    #[must_use]
+    pub fn stdout<W>() -> writer::Normalize<W, Self> {
+        Self::new(io::stdout(), Coloring::Auto, Verbosity::Default)
+    }
+}
+
+impl<Out: io::Write> Basic<Out> {
+    /// Creates a new [`Normalized`] [`Basic`] [`Writer`] outputting to the
+    /// given `output`.
+    ///
+    /// [`Normalized`]: writer::Normalized
+    #[must_use]
+    pub fn new<W>(
+        output: Out,
+        color: Coloring,
+        verbosity: impl Into<Verbosity>,
+    ) -> writer::Normalize<W, Self> {
+        Self::raw(output, color, verbosity).normalized()
+    }
+
+    /// Creates a new non-[`Normalized`] [`Basic`] [`Writer`] outputting to the
+    /// given `output`.
+    ///
+    /// Use it only if you know what you're doing. Otherwise, consider using
+    /// [`Basic::new()`] which creates an already [`Normalized`] version of a
+    /// [`Basic`] [`Writer`].
+    ///
+    /// [`Normalized`]: writer::Normalized
+    #[must_use]
+    pub fn raw(
+        output: Out,
+        color: Coloring,
+        verbosity: impl Into<Verbosity>,
+    ) -> Self {
+        let mut basic = Self {
+            output,
             styles: Styles::new(),
             indent: 0,
             lines_to_clear: 0,
-        }
-    }
-}
-
-impl Deref for Basic {
-    type Target = Term;
-
-    fn deref(&self) -> &Self::Target {
-        &self.terminal
-    }
-}
-
-impl Basic {
-    /// Creates a new [`Basic`] [`Writer`].
-    #[must_use]
-    pub fn new() -> Self {
-        Self::default()
+            verbosity: verbosity.into(),
+        };
+        basic.apply_cli(Cli {
+            verbose: u8::from(basic.verbosity) + 1,
+            color,
+        });
+        basic
     }
 
-    /// Clears last `n` lines if terminal is present.
+    /// Applies the given [`Cli`] options to this [`Basic`] [`Writer`].
+    pub fn apply_cli(&mut self, cli: Cli) {
+        match cli.verbose {
+            0 => {}
+            1 => self.verbosity = Verbosity::Default,
+            2 => self.verbosity = Verbosity::ShowWorld,
+            _ => self.verbosity = Verbosity::ShowWorldAndDocString,
+        };
+        self.styles.apply_coloring(cli.color);
+    }
+
+    /// Clears last `n` lines if [`Coloring`] is enabled.
     fn clear_last_lines_if_term_present(&mut self) -> io::Result<()> {
         if self.styles.is_present && self.lines_to_clear > 0 {
-            self.clear_last_lines(self.lines_to_clear)?;
+            self.output.clear_last_lines(self.lines_to_clear)?;
             self.lines_to_clear = 0;
         }
         Ok(())
@@ -184,38 +245,41 @@ impl Basic {
     /// Outputs the parsing `error` encountered while parsing some [`Feature`].
     ///
     /// [`Feature`]: gherkin::Feature
-    fn parsing_failed(&self, error: impl Display) -> io::Result<()> {
-        self.write_line(&self.styles.err(format!("Failed to parse: {}", error)))
+    pub(crate) fn parsing_failed(
+        &mut self,
+        error: impl Display,
+    ) -> io::Result<()> {
+        self.output
+            .write_line(&self.styles.err(format!("Failed to parse: {}", error)))
     }
 
-    /// Outputs [started] [`Feature`] to STDOUT.
+    /// Outputs the [started] [`Feature`].
     ///
     /// [started]: event::Feature::Started
     /// [`Feature`]: gherkin::Feature
-    fn feature_started(
+    pub(crate) fn feature_started(
         &mut self,
         feature: &gherkin::Feature,
     ) -> io::Result<()> {
         self.lines_to_clear = 1;
-        self.write_line(
+        self.output.write_line(
             &self
                 .styles
                 .ok(format!("{}: {}", feature.keyword, feature.name)),
         )
     }
 
-    /// Outputs [`Rule`] [started]/[scenario]/[finished] event to STDOUT.
+    /// Outputs the [`Rule`]'s [started]/[scenario]/[finished] event.
     ///
     /// [finished]: event::Rule::Finished
     /// [scenario]: event::Rule::Scenario
     /// [started]: event::Rule::Started
     /// [`Rule`]: gherkin::Rule
-    fn rule<W: Debug>(
+    pub(crate) fn rule<W: Debug>(
         &mut self,
         feat: &gherkin::Feature,
         rule: &gherkin::Rule,
         ev: event::Rule<W>,
-        cli: Cli,
     ) -> io::Result<()> {
         use event::Rule;
 
@@ -224,7 +288,7 @@ impl Basic {
                 self.rule_started(rule)?;
             }
             Rule::Scenario(sc, ev) => {
-                self.scenario(feat, &sc, &ev, cli)?;
+                self.scenario(feat, &sc, &ev)?;
             }
             Rule::Finished => {
                 self.indent = self.indent.saturating_sub(2);
@@ -233,14 +297,17 @@ impl Basic {
         Ok(())
     }
 
-    /// Outputs [started] [`Rule`] to STDOUT.
+    /// Outputs the [started] [`Rule`].
     ///
     /// [started]: event::Rule::Started
     /// [`Rule`]: gherkin::Rule
-    fn rule_started(&mut self, rule: &gherkin::Rule) -> io::Result<()> {
+    pub(crate) fn rule_started(
+        &mut self,
+        rule: &gherkin::Rule,
+    ) -> io::Result<()> {
         self.lines_to_clear = 1;
         self.indent += 2;
-        self.write_line(&self.styles.ok(format!(
+        self.output.write_line(&self.styles.ok(format!(
             "{indent}{}: {}",
             rule.keyword,
             rule.name,
@@ -248,18 +315,17 @@ impl Basic {
         )))
     }
 
-    /// Outputs [`Scenario`] [started]/[background]/[step] event to STDOUT.
+    /// Outputs the [`Scenario`]'s [started]/[background]/[step] event.
     ///
     /// [background]: event::Scenario::Background
     /// [started]: event::Scenario::Started
     /// [step]: event::Step
     /// [`Scenario`]: gherkin::Scenario
-    fn scenario<W: Debug>(
+    pub(crate) fn scenario<W: Debug>(
         &mut self,
         feat: &gherkin::Feature,
         scenario: &gherkin::Scenario,
         ev: &event::Scenario<W>,
-        cli: Cli,
     ) -> io::Result<()> {
         use event::{Hook, Scenario};
 
@@ -278,21 +344,21 @@ impl Basic {
                 self.indent = self.indent.saturating_sub(4);
             }
             Scenario::Background(bg, ev) => {
-                self.background(feat, bg, ev, cli)?;
+                self.background(feat, bg, ev)?;
             }
             Scenario::Step(st, ev) => {
-                self.step(feat, st, ev, cli)?;
+                self.step(feat, st, ev)?;
             }
             Scenario::Finished => self.indent = self.indent.saturating_sub(2),
         }
         Ok(())
     }
 
-    /// Outputs [failed] [`Scenario`]'s hook to STDOUT.
+    /// Outputs the [failed] [`Scenario`]'s hook.
     ///
     /// [failed]: event::Hook::Failed
     /// [`Scenario`]: gherkin::Scenario
-    fn hook_failed<W: Debug>(
+    pub(crate) fn hook_failed<W: Debug>(
         &mut self,
         feat: &gherkin::Feature,
         sc: &gherkin::Scenario,
@@ -302,7 +368,7 @@ impl Basic {
     ) -> io::Result<()> {
         self.clear_last_lines_if_term_present()?;
 
-        self.write_line(&self.styles.err(format!(
+        self.output.write_line(&self.styles.err(format!(
             "{indent}\u{2718}  Scenario's {} hook failed {}:{}:{}\n\
              {indent}   Captured output: {}{}",
             which,
@@ -323,17 +389,17 @@ impl Basic {
         )))
     }
 
-    /// Outputs [started] [`Scenario`] to STDOUT.
+    /// Outputs the [started] [`Scenario`].
     ///
     /// [started]: event::Scenario::Started
     /// [`Scenario`]: gherkin::Scenario
-    fn scenario_started(
+    pub(crate) fn scenario_started(
         &mut self,
         scenario: &gherkin::Scenario,
     ) -> io::Result<()> {
         self.lines_to_clear = 1;
         self.indent += 2;
-        self.write_line(&self.styles.ok(format!(
+        self.output.write_line(&self.styles.ok(format!(
             "{}{}: {}",
             " ".repeat(self.indent),
             scenario.keyword,
@@ -341,45 +407,44 @@ impl Basic {
         )))
     }
 
-    /// Outputs [`Step`] [started]/[passed]/[skipped]/[failed] event to STDOUT.
+    /// Outputs the [`Step`]'s [started]/[passed]/[skipped]/[failed] event.
     ///
     /// [failed]: event::Step::Failed
     /// [passed]: event::Step::Passed
     /// [skipped]: event::Step::Skipped
     /// [started]: event::Step::Started
     /// [`Step`]: gherkin::Step
-    fn step<W: Debug>(
+    pub(crate) fn step<W: Debug>(
         &mut self,
         feat: &gherkin::Feature,
         step: &gherkin::Step,
         ev: &event::Step<W>,
-        cli: Cli,
     ) -> io::Result<()> {
         use event::Step;
 
         match ev {
             Step::Started => {
-                self.step_started(step, cli)?;
+                self.step_started(step)?;
             }
             Step::Passed(captures) => {
-                self.step_passed(step, captures, cli)?;
+                self.step_passed(step, captures)?;
                 self.indent = self.indent.saturating_sub(4);
             }
             Step::Skipped => {
-                self.step_skipped(feat, step, cli)?;
+                self.step_skipped(feat, step)?;
                 self.indent = self.indent.saturating_sub(4);
             }
             Step::Failed(c, w, i) => {
-                self.step_failed(feat, step, c.as_ref(), w.as_ref(), i, cli)?;
+                self.step_failed(feat, step, c.as_ref(), w.as_ref(), i)?;
                 self.indent = self.indent.saturating_sub(4);
             }
         }
         Ok(())
     }
 
-    /// Outputs [started] [`Step`] to STDOUT.
+    /// Outputs the [started] [`Step`].
     ///
-    /// This [`Step`] is printed only if terminal is present and gets
+    /// The [`Step`] is printed only if [`Coloring`] is enabled and gets
     /// overwritten by later [passed]/[skipped]/[failed] events.
     ///
     /// [failed]: event::Step::Failed
@@ -387,10 +452,9 @@ impl Basic {
     /// [skipped]: event::Step::Skipped
     /// [started]: event::Step::Started
     /// [`Step`]: gherkin::Step
-    fn step_started(
+    pub(crate) fn step_started(
         &mut self,
         step: &gherkin::Step,
-        cli: Cli,
     ) -> io::Result<()> {
         self.indent += 4;
         if self.styles.is_present {
@@ -400,11 +464,13 @@ impl Basic {
                 step.value,
                 step.docstring
                     .as_ref()
-                    .and_then(|doc| cli.verbose.then(
-                        || format_str_with_indent(
-                            doc,
-                            self.indent.saturating_sub(3) + 3,
-                        )
+                    .and_then(|doc| self.verbosity.shows_docstring().then(
+                        || {
+                            format_str_with_indent(
+                                doc,
+                                self.indent.saturating_sub(3) + 3,
+                            )
+                        }
                     ))
                     .unwrap_or_default(),
                 step.table
@@ -419,15 +485,14 @@ impl Basic {
         Ok(())
     }
 
-    /// Outputs [passed] [`Step`] to STDOUT.
+    /// Outputs the [passed] [`Step`].
     ///
     /// [passed]: event::Step::Passed
     /// [`Step`]: gherkin::Step
-    fn step_passed(
+    pub(crate) fn step_passed(
         &mut self,
         step: &gherkin::Step,
         captures: &CaptureLocations,
-        cli: Cli,
     ) -> io::Result<()> {
         self.clear_last_lines_if_term_present()?;
 
@@ -443,7 +508,7 @@ impl Basic {
             .docstring
             .as_ref()
             .and_then(|doc| {
-                cli.verbose.then(|| {
+                self.verbosity.shows_docstring().then(|| {
                     format_str_with_indent(
                         doc,
                         self.indent.saturating_sub(3) + 3,
@@ -457,7 +522,7 @@ impl Basic {
             .map(|t| format_table(t, self.indent))
             .unwrap_or_default());
 
-        self.write_line(&self.styles.ok(format!(
+        self.output.write_line(&self.styles.ok(format!(
             "{indent}{} {}{}{}",
             step_keyword,
             step_value,
@@ -467,28 +532,29 @@ impl Basic {
         )))
     }
 
-    /// Outputs [skipped] [`Step`] to STDOUT.
+    /// Outputs the [skipped] [`Step`].
     ///
     /// [skipped]: event::Step::Skipped
     /// [`Step`]: gherkin::Step
-    fn step_skipped(
+    pub(crate) fn step_skipped(
         &mut self,
         feat: &gherkin::Feature,
         step: &gherkin::Step,
-        cli: Cli,
     ) -> io::Result<()> {
         self.clear_last_lines_if_term_present()?;
-        self.write_line(&self.styles.skipped(format!(
+        self.output.write_line(&self.styles.skipped(format!(
             "{indent}?  {} {}{}{}\n\
              {indent}   Step skipped: {}:{}:{}",
             step.keyword,
             step.value,
             step.docstring
                 .as_ref()
-                .and_then(|doc| cli.verbose.then(|| format_str_with_indent(
-                    doc,
-                    self.indent.saturating_sub(3) + 3,
-                )))
+                .and_then(|doc| self.verbosity.shows_docstring().then(|| {
+                    format_str_with_indent(
+                        doc,
+                        self.indent.saturating_sub(3) + 3,
+                    )
+                }))
                 .unwrap_or_default(),
             step.table
                 .as_ref()
@@ -504,18 +570,17 @@ impl Basic {
         )))
     }
 
-    /// Outputs [failed] [`Step`] to STDOUT.
+    /// Outputs the [failed] [`Step`].
     ///
     /// [failed]: event::Step::Failed
     /// [`Step`]: gherkin::Step
-    fn step_failed<W: Debug>(
+    pub(crate) fn step_failed<W: Debug>(
         &mut self,
         feat: &gherkin::Feature,
         step: &gherkin::Step,
         captures: Option<&CaptureLocations>,
         world: Option<&W>,
         err: &event::StepError,
-        cli: Cli,
     ) -> io::Result<()> {
         self.clear_last_lines_if_term_present()?;
 
@@ -526,10 +591,10 @@ impl Basic {
         ));
         let step_value = captures.map_or_else(
             || self.styles.err(&step.value),
-            |captures| {
+            |capts| {
                 format_captures(
                     &step.value,
-                    captures,
+                    capts,
                     |v| self.styles.err(v),
                     |v| self.styles.err(self.styles.bold(v)),
                 )
@@ -543,10 +608,12 @@ impl Basic {
              {indent}   Captured output: {}{}",
             step.docstring
                 .as_ref()
-                .and_then(|doc| cli.verbose.then(|| format_str_with_indent(
-                    doc,
-                    self.indent.saturating_sub(3) + 3,
-                )))
+                .and_then(|doc| self.verbosity.shows_docstring().then(|| {
+                    format_str_with_indent(
+                        doc,
+                        self.indent.saturating_sub(3) + 3,
+                    )
+                }))
                 .unwrap_or_default(),
             step.table
                 .as_ref()
@@ -567,6 +634,7 @@ impl Basic {
                     format!("{:#?}", w),
                     self.indent.saturating_sub(3) + 3,
                 ))
+                .filter(|_| self.verbosity.shows_world())
                 .unwrap_or_default(),
             indent = " ".repeat(self.indent.saturating_sub(3))
         ));
@@ -577,8 +645,8 @@ impl Basic {
         ))
     }
 
-    /// Outputs [`Background`] [`Step`] [started]/[passed]/[skipped]/[failed]
-    /// event to STDOUT.
+    /// Outputs the [`Background`] [`Step`]'s
+    /// [started]/[passed]/[skipped]/[failed] event.
     ///
     /// [failed]: event::Step::Failed
     /// [passed]: event::Step::Passed
@@ -586,38 +654,37 @@ impl Basic {
     /// [started]: event::Step::Started
     /// [`Background`]: gherkin::Background
     /// [`Step`]: gherkin::Step
-    fn background<W: Debug>(
+    pub(crate) fn background<W: Debug>(
         &mut self,
         feat: &gherkin::Feature,
         bg: &gherkin::Step,
         ev: &event::Step<W>,
-        cli: Cli,
     ) -> io::Result<()> {
         use event::Step;
 
         match ev {
             Step::Started => {
-                self.bg_step_started(bg, cli)?;
+                self.bg_step_started(bg)?;
             }
             Step::Passed(captures) => {
-                self.bg_step_passed(bg, captures, cli)?;
+                self.bg_step_passed(bg, captures)?;
                 self.indent = self.indent.saturating_sub(4);
             }
             Step::Skipped => {
-                self.bg_step_skipped(feat, bg, cli)?;
+                self.bg_step_skipped(feat, bg)?;
                 self.indent = self.indent.saturating_sub(4);
             }
             Step::Failed(c, w, i) => {
-                self.bg_step_failed(feat, bg, c.as_ref(), w.as_ref(), i, cli)?;
+                self.bg_step_failed(feat, bg, c.as_ref(), w.as_ref(), i)?;
                 self.indent = self.indent.saturating_sub(4);
             }
         }
         Ok(())
     }
 
-    /// Outputs [started] [`Background`] [`Step`] to STDOUT.
+    /// Outputs the [started] [`Background`] [`Step`].
     ///
-    /// This [`Step`] is printed only if terminal is present and gets
+    /// The [`Step`] is printed only if [`Coloring`] is enabled and gets
     /// overwritten by later [passed]/[skipped]/[failed] events.
     ///
     /// [failed]: event::Step::Failed
@@ -626,10 +693,9 @@ impl Basic {
     /// [started]: event::Step::Started
     /// [`Background`]: gherkin::Background
     /// [`Step`]: gherkin::Step
-    fn bg_step_started(
+    pub(crate) fn bg_step_started(
         &mut self,
         step: &gherkin::Step,
-        cli: Cli,
     ) -> io::Result<()> {
         self.indent += 4;
         if self.styles.is_present {
@@ -639,11 +705,13 @@ impl Basic {
                 step.value,
                 step.docstring
                     .as_ref()
-                    .and_then(|doc| cli.verbose.then(
-                        || format_str_with_indent(
-                            doc,
-                            self.indent.saturating_sub(3) + 3,
-                        )
+                    .and_then(|doc| self.verbosity.shows_docstring().then(
+                        || {
+                            format_str_with_indent(
+                                doc,
+                                self.indent.saturating_sub(3) + 3,
+                            )
+                        }
                     ))
                     .unwrap_or_default(),
                 step.table
@@ -658,16 +726,15 @@ impl Basic {
         Ok(())
     }
 
-    /// Outputs [passed] [`Background`] [`Step`] to STDOUT.
+    /// Outputs the [passed] [`Background`] [`Step`].
     ///
     /// [passed]: event::Step::Passed
     /// [`Background`]: gherkin::Background
     /// [`Step`]: gherkin::Step
-    fn bg_step_passed(
+    pub(crate) fn bg_step_passed(
         &mut self,
         step: &gherkin::Step,
         captures: &CaptureLocations,
-        cli: Cli,
     ) -> io::Result<()> {
         self.clear_last_lines_if_term_present()?;
 
@@ -683,7 +750,7 @@ impl Basic {
             .docstring
             .as_ref()
             .and_then(|doc| {
-                cli.verbose.then(|| {
+                self.verbosity.shows_docstring().then(|| {
                     format_str_with_indent(
                         doc,
                         self.indent.saturating_sub(3) + 3,
@@ -697,7 +764,7 @@ impl Basic {
             .map(|t| format_table(t, self.indent))
             .unwrap_or_default());
 
-        self.write_line(&self.styles.ok(format!(
+        self.output.write_line(&self.styles.ok(format!(
             "{indent}{} {}{}{}",
             step_keyword,
             step_value,
@@ -707,29 +774,30 @@ impl Basic {
         )))
     }
 
-    /// Outputs [skipped] [`Background`] [`Step`] to STDOUT.
+    /// Outputs the [skipped] [`Background`] [`Step`].
     ///
     /// [skipped]: event::Step::Skipped
     /// [`Background`]: gherkin::Background
     /// [`Step`]: gherkin::Step
-    fn bg_step_skipped(
+    pub(crate) fn bg_step_skipped(
         &mut self,
         feat: &gherkin::Feature,
         step: &gherkin::Step,
-        cli: Cli,
     ) -> io::Result<()> {
         self.clear_last_lines_if_term_present()?;
-        self.write_line(&self.styles.skipped(format!(
+        self.output.write_line(&self.styles.skipped(format!(
             "{indent}?> {} {}{}{}\n\
              {indent}   Background step failed: {}:{}:{}",
             step.keyword,
             step.value,
             step.docstring
                 .as_ref()
-                .and_then(|doc| cli.verbose.then(|| format_str_with_indent(
-                    doc,
-                    self.indent.saturating_sub(3) + 3,
-                )))
+                .and_then(|doc| self.verbosity.shows_docstring().then(|| {
+                    format_str_with_indent(
+                        doc,
+                        self.indent.saturating_sub(3) + 3,
+                    )
+                }))
                 .unwrap_or_default(),
             step.table
                 .as_ref()
@@ -745,19 +813,18 @@ impl Basic {
         )))
     }
 
-    /// Outputs [failed] [`Background`] [`Step`] to STDOUT.
+    /// Outputs the [failed] [`Background`] [`Step`].
     ///
     /// [failed]: event::Step::Failed
     /// [`Background`]: gherkin::Background
     /// [`Step`]: gherkin::Step
-    fn bg_step_failed<W: Debug>(
+    pub(crate) fn bg_step_failed<W: Debug>(
         &mut self,
         feat: &gherkin::Feature,
         step: &gherkin::Step,
         captures: Option<&CaptureLocations>,
         world: Option<&W>,
         err: &event::StepError,
-        cli: Cli,
     ) -> io::Result<()> {
         self.clear_last_lines_if_term_present()?;
 
@@ -768,10 +835,10 @@ impl Basic {
         ));
         let step_value = captures.map_or_else(
             || self.styles.err(&step.value),
-            |captures| {
+            |capts| {
                 format_captures(
                     &step.value,
-                    captures,
+                    capts,
                     |v| self.styles.err(v),
                     |v| self.styles.err(self.styles.bold(v)),
                 )
@@ -785,10 +852,12 @@ impl Basic {
              {indent}   Captured output: {}{}",
             step.docstring
                 .as_ref()
-                .and_then(|doc| cli.verbose.then(|| format_str_with_indent(
-                    doc,
-                    self.indent.saturating_sub(3) + 3,
-                )))
+                .and_then(|doc| self.verbosity.shows_docstring().then(|| {
+                    format_str_with_indent(
+                        doc,
+                        self.indent.saturating_sub(3) + 3,
+                    )
+                }))
                 .unwrap_or_default(),
             step.table
                 .as_ref()
@@ -854,8 +923,8 @@ fn format_table(table: &gherkin::Table, indent: usize) -> String {
         .rows
         .iter()
         .fold(None, |mut acc: Option<Vec<_>>, row| {
-            if let Some(acc) = acc.as_mut() {
-                for (cell, max_len) in row.iter().zip(acc) {
+            if let Some(existing_len) = acc.as_mut() {
+                for (cell, max_len) in row.iter().zip(existing_len) {
                     *max_len = cmp::max(*max_len, cell.len());
                 }
             } else {
