@@ -1,4 +1,4 @@
-// Copyright (c) 2018-2021  Brendan Molloy <brendan@bbqsrc.net>,
+// Copyright (c) 2018-2022  Brendan Molloy <brendan@bbqsrc.net>,
 //                          Ilya Solovyiov <ilya.solovyiov@gmail.com>,
 //                          Kai Ren <tyranron@gmail.com>
 //
@@ -19,9 +19,7 @@ use std::sync::Arc;
 use async_trait::async_trait;
 use derive_more::Deref;
 
-use crate::{
-    event, parser, ArbitraryWriter, Event, FailureWriter, World, Writer,
-};
+use crate::{event, parser, writer, Event, World, Writer};
 
 /// [`Writer`]-wrapper for transforming [`Skipped`] [`Step`]s into [`Failed`].
 ///
@@ -59,13 +57,13 @@ where
         Option<&gherkin::Rule>,
         &gherkin::Scenario,
     ) -> bool,
-    Wr: for<'val> ArbitraryWriter<'val, W, String>,
+    Wr: for<'val> writer::Arbitrary<'val, W, String>,
 {
     type Cli = Wr::Cli;
 
     async fn handle_event(
         &mut self,
-        ev: parser::Result<Event<event::Cucumber<W>>>,
+        event: parser::Result<Event<event::Cucumber<W>>>,
         cli: &Self::Cli,
     ) {
         use event::{
@@ -73,17 +71,17 @@ where
         };
 
         let map_failed = |f: Arc<_>, r: Option<Arc<_>>, sc: Arc<_>, st: _| {
-            let event = if (self.should_fail)(&f, r.as_deref(), &sc) {
+            let ev = if (self.should_fail)(&f, r.as_deref(), &sc) {
                 Step::Failed(None, None, Panic(Arc::new("not allowed to skip")))
             } else {
                 Step::Skipped
             };
 
-            Cucumber::scenario(f, r, sc, Scenario::Step(st, event))
+            Cucumber::scenario(f, r, sc, Scenario::Step(st, ev))
         };
 
-        let ev = ev.map(|ev| {
-            ev.map(|ev| match ev {
+        let event = event.map(|outer| {
+            outer.map(|ev| match ev {
                 Cucumber::Feature(
                     f,
                     Feature::Rule(
@@ -101,16 +99,17 @@ where
             })
         });
 
-        self.writer.handle_event(ev, cli).await;
+        self.writer.handle_event(event, cli).await;
     }
 }
 
 #[async_trait(?Send)]
-impl<'val, W, Wr, Val, F> ArbitraryWriter<'val, W, Val> for FailOnSkipped<Wr, F>
+impl<'val, W, Wr, Val, F> writer::Arbitrary<'val, W, Val>
+    for FailOnSkipped<Wr, F>
 where
     W: World,
     Self: Writer<W>,
-    Wr: ArbitraryWriter<'val, W, Val>,
+    Wr: writer::Arbitrary<'val, W, Val>,
     Val: 'val,
 {
     async fn write(&mut self, val: Val)
@@ -121,9 +120,9 @@ where
     }
 }
 
-impl<W, Wr, F> FailureWriter<W> for FailOnSkipped<Wr, F>
+impl<W, Wr, F> writer::Failure<W> for FailOnSkipped<Wr, F>
 where
-    Wr: FailureWriter<W>,
+    Wr: writer::Failure<W>,
     Self: Writer<W>,
 {
     fn failed_steps(&self) -> usize {
@@ -139,12 +138,14 @@ where
     }
 }
 
+impl<Wr: writer::Normalized, F> writer::Normalized for FailOnSkipped<Wr, F> {}
+
 impl<Writer> From<Writer> for FailOnSkipped<Writer> {
     fn from(writer: Writer) -> Self {
         Self {
             writer,
             should_fail: |_, _, sc| {
-                !sc.tags.iter().any(|tag| tag == "allow_skipped")
+                !sc.tags.iter().any(|t| t == "allow.skipped")
             },
         }
     }
